/* FILE: diag.dsp               Copyright (c), Symmetric Research, 1992
 *
 * DSP32C portion of the diagnostics.
 *
 * NOTE:
 *
 * The diagnostics have been located in the internal DSP32C ram block in
 * high memory.  This location avoids being overwritten as the diagnostics
 * proceed, and also avoids the area used by the DSPMON break code, allowing
 * for single stepping if necessary.
 *
 *
 */

constant

        DIAGNOSTICS     = 0xFFFC00,     /* base address for diagnostics */

        COMPUTING       = 0,            /* status flag values */
        SUCCESS         = 1,
        ERROR           = 2;



<DIAGNOSTICS>

long input_array[5] = { 0 };

slam {

        /* Test for maximum current draw and data bus ringing. */

        struct input_array

                { long status, address, nbytes, waits; } input;

        register

                i               = r1,
                p               = r2,
                ones            = r3,
                zero            = r4,
                waits           = r5,
                nbytes          = r6,
                status          = r7,
                address         = r8,
                temp_ones       = r9,
                temp_zero       = r10;



        /* Set the waits states. */

        p = input.waits; waits = *p; pcw = waits;


        /* Slam all ones followed by all zeros onto the data bus. */

        address = input.address; address = *address;

                nbytes = input.nbytes; nbytes = *nbytes;

                        ones = 0x00FFFFFF; zero = 0x00000000;

        p = address;

            nbytes &~= 0x7;  /* chop nbytes to a multiple of 8 */


        for ( i = 0 ; i < nbytes ; i += 8 )

                { *p++ = ones; *p++ = zero; }




        /* Read them back, checking for an error. */

        p = address;

        for ( i = 0 , status = SUCCESS ; i < nbytes ; i += 8 ) {

                temp_ones = *p++; temp_zero = *p++;

                    if ( temp_ones != ones )

                                { status = ERROR; p -= 8; goto slam_error; }

                    if ( temp_zero != zero )

                                { status = ERROR; p -= 8; goto slam_error; }

                } slam_error:

          address = p;



        /* Set the return values. */

        p = input.status; *p = status;

          p = input.address; *p = address;



        /* Wait for the PC. */

        slam_wait: goto slam_wait;


}
slide {


        /* Perform a sliding ones memory test. */

        struct input_array

                { long status, address, nbytes, waits; } input;

        register

                i               = r1,
                p               = r2,
                next            = r3,
                ones            = r4,
                value           = r5,
                waits           = r6,
                status          = r7,
                nbytes          = r8,
                address         = r9;



        /* Set the waits states. */

        p = input.waits; waits = *p; pcw = waits;



        /* Slide, one byte at a time. */

        address = input.address; p = *address;

                nbytes = input.nbytes; nbytes = *nbytes;

                        ones = 0x00FFFFFF;


        for ( i = 0 , status = SUCCESS ; i < nbytes ; i++ ) {


    /* Slide value to the right, charging the bus between reads/writes. */

        value = 1; *p = value.b; next = p+4; *next = ones.b;

    value = *p.b; *next = ones.b; value *= 2; *p = value.b; *next = ones.b;
    value = *p.b; *next = ones.b; value *= 2; *p = value.b; *next = ones.b;
    value = *p.b; *next = ones.b; value *= 2; *p = value.b; *next = ones.b;
    value = *p.b; *next = ones.b; value *= 2; *p = value.b; *next = ones.b;
    value = *p.b; *next = ones.b; value *= 2; *p = value.b; *next = ones.b;
    value = *p.b; *next = ones.b; value *= 2; *p = value.b; *next = ones.b;
    value = *p.b; *next = ones.b; value *= 2; *p = value.b; *next = ones.b;


                /* Check the value, returning an error if not right. */

                if ( value != 0x80 )

                        { status = ERROR; goto slide_error; }


                /* Increment on to the next byte. */

                p++;

                } slide_error:


          address = p;



        /* Set the return values. */

        p = input.status; *p = status;

          p = input.address; *p = address;



        /* Wait for the PC. */

        slide_wait: goto slide_wait;

}
