/* FILE: output.c               Copyright (c), Symmetric Research, 1998-2000
 *
 * Output functions and emits for DSPASM.
 *
 *
 *
 *
 */

#include <math.h>       /* Needed for labs */

#include "dspasm.h"





/* OUTPUT BUFFER FUNCTIONS:
 *
 *
 *
 */

#define MAXBYTES        (unsigned)0xFFF0        /* output buffer size   */

char
        SRFAR *output_buffer,                   /* buffer start         */
        SRFAR *output;                          /* current position     */


unsigned

        nbytes = 0;                             /* current total bytes  */



#define END_RECORD      0xFFFFFFFFL




int allocate_output_buffer( void) {

        /* Allocate an output buffer. */

        output_buffer = SrAlloc( MAXBYTES);


        if ( output_buffer )

                { output = output_buffer; return( 1); }

        else
                return( 0);

}





/* OUTPUT FILE FUNCTIONS:
 *
 *
 *
 *
 */

char *newext( char *filename, char *ext) {

        char *p, *strchr(), *strcat();


        /* Change the extension on the filename. */

        if ( (p = strchr( filename, '.')) != NULL )

                *p = '\0';
        

        return( strcat( filename, ext));

}       
void write_dso( void) {

        SRFILEHANDLE output_handle; int i; SYMBOL *s;

          struct {

                 unsigned short name_length;
                 unsigned long  value;
                 unsigned short type;
                          
                } o;




        /* Dump the output buffer in .dso binary format. */

        newext( filename, ".dso");

            printf( "\t\t writing %s\n", filename);


        output_handle = SrCreate( filename);
              SrWrite( output_handle, output_buffer, nbytes);




        /* Add the global symbols onto the end of the file. */

        for ( i = 0 ; i < NHASH ; i++ ) {

                for ( s = gsymbol[i] ; s ; s = s->next ) {

                        o.name_length   = strlen( s->name)+1;
                        o.value         = s->value.l;
                        o.type          = s->type;

                        SrWrite( output_handle, &o, sizeof( o));
                        SrWrite( output_handle, s->name, o.name_length);

                        }
                }



        /* Close. */

        SrClose( output_handle);


}
void write_dsh( void) {

        FILE *fp; char array_name[32];

        static char *fmt[] = {

                "/* hexified output from %s */\n",
                "\n",
                "unsigned long %s[] = {\n",
                "\n",
                "               0x%08lX,\n",
                "\n",
                "       };\n",

                };




        /* Dump the output buffer in hexified C include file format. */

        newext( filename, ".dsh");

                printf( "\t\t writing %s\n", filename);

                fp = fopen( filename, "w");

                fprintf( fp, fmt[0], newext( filename, ".dsp"));
                fprintf( fp, fmt[1]);

        strcpy( array_name, filename);
        
                newext( array_name, "_dsh");

                fprintf( fp, fmt[2], array_name);
                fprintf( fp, fmt[3]);


        for ( output = output_buffer ; nbytes ; nbytes -= 4 )

                fprintf( fp, fmt[4], *((unsigned long SRFAR *)output)++);


        fprintf( fp, fmt[5]);
        fprintf( fp, fmt[6]);
        
        fclose( fp);


}
void write_dsl( void) {
        
        FILE *fp; unsigned long SRFAR *o, start, nbytes, value;

            char buffer[80];

        static char *fmt[] = {

                "Listing file from %s\n",
                "\n",
                " Record at %06lX, nbytes = 0x%08lX\n",
                "\n",
                "  %06lX:[%08lX] %s\n",
                "\n",

                };



        /* Write an ASCII listing in AT&T mnemonics. */

        newext( filename, ".dsl");

                printf( "\t\t writing %s\n", filename);
                
                fp = fopen( filename, "w");

                fprintf( fp, fmt[0], newext( filename, ".dsp"));
                fprintf( fp, fmt[1]);



        /* Point to the output buffer and generate the listing. */
        
        o = (unsigned long SRFAR *)output_buffer;

                /* skip over leading SR_ID */ o++;


        while ( *o != END_RECORD ) {

                start = *o++;
                nbytes = *o++;

                fprintf( fp, fmt[2], start, nbytes);
                fprintf( fp, fmt[3]);

                while ( nbytes ) { value = *o++;
                        
                    DSP32C_Decode( value, buffer);

                        fprintf( fp, fmt[4], start, value, buffer);

                                nbytes -= 4; start += 4;
                        }

                fprintf( fp, fmt[5]);

                }

        fclose( fp);

}
char *typename( unsigned type) {

        static struct { unsigned type; char *typename; }

            *t, typename_table[] = {

                CONS_SYM,       "CONS (0001)",
                BYTE_SYM,       "BYTE (0002)",
                WORD_SYM,       "WORD (0004)",
                LONG_SYM,       "LONG (0008)",
                DSPF_SYM,       "DSPF (0010)",
                IEEE_SYM,       "IEEE (0020)",
                FUNC_SYM,       "FUNC (0040)",
                GOTO_SYM,       "GOTO (0080)",

                0,              "  NO TYPE  ",

                };


        /* Convert a symbol type to a string for the symbol listing. */

        for ( t = typename_table ; t->type ; t++ ) {
                
                if ( t->type == type )

                        return( t->typename);

                }

        return( t->typename);

}
void write_sym( unsigned format) {

        FILE *fp; SYMBOL *s; int i;
        
        static struct { char *ext, *fmt[5]; } *t, table[3] = {
                    
                ".dss",                                 /* format 0 */

                        "Global symbol values from %s\n",
                        "\n",
                        "  %-25s     %08lX    %s\n",
                        "\n",
                        "\n",


                ".dsa",                                 /* format 1 */
                    
                        "constant /* global symbol values from %s */\n",
                        "\n",
                        "  %-25s     = %08lX,    /* %s */\n",
                        "\n",
                        "       ;\n",


                ".dsc",                                 /* format 2 */
                    
                        "/* global symbol values from %s */\n",
                        "\n",
                        "#define DSP_%-25s     0x%08lXL     /* %s */\n",
                        "\n",
                        "\n",

                        };




    /* Dump the global symbols in the requested ASCII format. */



        /* Point to the right set of format statements. */

        t = &table[format];




        /* Dump the symbol table. */

        newext( filename, t->ext);

            printf( "\t\t writing %s\n", filename);
            
            fp = fopen( filename, "w");

                fprintf( fp, t->fmt[0], newext( filename, ".dsp"));
                fprintf( fp, t->fmt[1]);


        for ( i = 0 ; i < NHASH ; i++ ) {
                
          for ( s = gsymbol[i] ; s ; s = s->next )
                        
             fprintf( fp, t->fmt[2], s->name, s->value.l, typename( s->type));

          }
        


        fprintf( fp, t->fmt[3]);
        fprintf( fp, t->fmt[4]);

        fclose( fp);

}







/* PROGRAM LOCATION COUNTER AND BASIC OUTPUT BUFFER EMITS:
 *
 * NOTE:
 *
 * These functions are in caps to make them easy to find when called from
 * dspasm.c.
 *
 *
 */

#define SR_ID           0x52535F5FL             /* = "__SR"             */


typedef struct record_header

        { unsigned long start, nbytes; } RECORD_HEADER;



RECORD_HEADER SRFAR *record = NULL;

unsigned long pc = 0L;




void FILE_ID( void) {


        /* Emit the leading .dso object file id. */
        
        *((unsigned long SRFAR *)output)++ = SR_ID;

          nbytes += 4;
        

}
void CHECK_RECORD( void) {


        /* Check to see if a new record needs to be started. */

        if ( !record || record->start+record->nbytes != pc ) {

                record = (RECORD_HEADER SRFAR *)output;

                record->start  = pc;
                record->nbytes = 0;

                output += sizeof( RECORD_HEADER);
                nbytes += sizeof( RECORD_HEADER);

                }


}
void FINAL_RECORD( void) {


        /* Emit a final record indicating end of output. */

        record = (RECORD_HEADER SRFAR *)output;

        record->start  = END_RECORD;
        record->nbytes = 0;

        output += sizeof( RECORD_HEADER);
        nbytes += sizeof( RECORD_HEADER);



}       
void emit( unsigned long value) {


        /* Emit an unsigned long to the current record. */
        
        nbytes += 4;

        if ( nbytes > MAXBYTES )

                { printf( "ERROR: output buffer overrun\n"); exit( 1); }


        *((unsigned long SRFAR *)output)++ = value;


        record->nbytes += 4;
                    pc += 4;


}       
void rewind_output( unsigned long *value) {


        /* Rewind in the current record, saving the rewound value. */

        nbytes -= 4;

          *value = *(--(unsigned long SRFAR *)output);


        record->nbytes -= 4;
                    pc -= 4;


}
void DATA( int type, unsigned long value) {



        /* Emit initialized data of the specified type to output buffer. */
        
        switch ( type ) {
                

                case BYTE:

                        *((unsigned char SRFAR *)output)++ = value;

                        break;


                case WORD:

                        *((unsigned short SRFAR *)output)++ = value;

                        break;
                        

                default:                        
        
                        *((unsigned long SRFAR *)output)++ = value;

                        break;

                }



        nbytes += datasize[ type];

        if ( nbytes > MAXBYTES )

                { printf( "ERROR: output buffer overrun\n"); exit( 1); }



        record->nbytes += datasize[ type];
                    pc += datasize[ type];



}
void reset_output_buffer( void) {


        /* Reset the output buffer for a second pass. */
        
        output = output_buffer;

                nbytes = 0; pc = 0;


}       






/* Helper function for checking the length of constants. */

int check( unsigned long *N, long max ) {

        if ( labs( *N ) > max )

                error( "constant is too large");

        *N &= max;

           return( 1); /* success */

}








/* CAU EMITS:
 *
 *
 *
 *
 */

static unsigned long


        H[] = {

                0x00L <<16,  /* 00  zero    */

                0x01L <<16,  /* 01  r1      */
                0x02L <<16,  /* 02  r2      */
                0x03L <<16,  /* 03  r3      */
                0x04L <<16,  /* 04  r4      */
                0x05L <<16,  /* 05  r5      */
                0x06L <<16,  /* 06  r6      */
                0x07L <<16,  /* 07  r7      */
                0x08L <<16,  /* 08  r8      */
                0x09L <<16,  /* 09  r9      */
                0x0AL <<16,  /* 10  r10     */
                0x0BL <<16,  /* 11  r11     */
                0x0CL <<16,  /* 12  r12     */
                0x0DL <<16,  /* 13  r13     */
                0x0EL <<16,  /* 14  r14     */
                0x11L <<16,  /* 15  r15     */
                0x12L <<16,  /* 16  r16     */
                0x13L <<16,  /* 17  r17     */
                0x14L <<16,  /* 18  r18     */
                0x15L <<16,  /* 19  r19     */
                0x18L <<16,  /* 20  r20     */
                0x19L <<16,  /* 21  r21     */
                0x1DL <<16,  /* 22  r22     */

                0x0FL <<16,  /* 23  pc      */
                0x1EL <<16,  /* 24  pcsh    */

                },



        F[] = {

                0x00L <<21,  /* 00  +     addition                 */
                0x01L <<21,  /* 01  *2    mult by 2                */
                0x02L <<21,  /* 02  -     subtract {N,rS} - rD     */
                0x03L <<21,  /* 03  #     reverse carry add        */
                0x04L <<21,  /* 04  -     subtract rD - {N,rS}     */
                0x05L <<21,  /* 05  -     negation                 */
                0x06L <<21,  /* 06  &~    and with complement      */
                0x07L <<21,  /* 07  -     compare w/o store        */
                0x08L <<21,  /* 08  ^     xor                      */
                0x09L <<21,  /* 09  >>>1  rotate right thru carry  */
                0x0AL <<21,  /* 10  |     bitwise or               */
                0x0BL <<21,  /* 11  <<<1  rotate left thru carry   */
                0x0CL <<21,  /* 12  >>1   shift right by 1         */
                0x0DL <<21,  /* 13  /2    divide by 2              */
                0x0EL <<21,  /* 14  &     bitwise and              */
                0x0FL <<21,  /* 15  &     bitwise and w/o store    */

                },



        W[] = {

                0x03L <<22,  /* 00  long   default  */
                0x01L <<22,  /* 01  byte            */
                0x02L <<22,  /* 02  word            */
                0x03L <<22,  /* 03  long            */

                },



        r[] = {

                0x00L <<10,  /* 00                    */
                0x01L <<10,  /* 01  bit reversed inc  */

                },



        P[] = {


                0x00L << 5,  /* 00  null    */

                0x01L << 5,  /* 01  r1      */
                0x02L << 5,  /* 02  r2      */
                0x03L << 5,  /* 03  r3      */
                0x04L << 5,  /* 04  r4      */
                0x05L << 5,  /* 05  r5      */
                0x06L << 5,  /* 06  r6      */
                0x07L << 5,  /* 07  r7      */
                0x08L << 5,  /* 08  r8      */
                0x09L << 5,  /* 09  r9      */
                0x0AL << 5,  /* 10  r10     */
                0x0BL << 5,  /* 11  r11     */
                0x0CL << 5,  /* 12  r12     */
                0x0DL << 5,  /* 13  r13     */
                0x0EL << 5,  /* 14  r14     */
                0x11L << 5,  /* 15  r15     */
                0x12L << 5,  /* 16  r16     */
                0x13L << 5,  /* 17  r17     */
                0x14L << 5,  /* 18  r18     */
                0x15L << 5,  /* 19  r19     */
                0x18L << 5,  /* 20  r20     */
                0x19L << 5,  /* 21  r21     */
                0x1DL << 5,  /* 22  r22     */

                },



        I[] = {

                0x00L     ,  /* 00  null    */

                0x01L     ,  /* 01  ++r1    */
                0x02L     ,  /* 02  ++r2    */
                0x03L     ,  /* 03  ++r3    */
                0x04L     ,  /* 04  ++r4    */
                0x05L     ,  /* 05  ++r5    */
                0x06L     ,  /* 06  ++r6    */
                0x07L     ,  /* 07  ++r7    */
                0x08L     ,  /* 08  ++r8    */
                0x09L     ,  /* 09  ++r9    */
                0x0AL     ,  /* 10  ++r10   */
                0x0BL     ,  /* 11  ++r11   */
                0x0CL     ,  /* 12  ++r12   */
                0x0DL     ,  /* 13  ++r13   */
                0x0EL     ,  /* 14  ++r14   */
                0x11L     ,  /* 15  ++r15   */
                0x12L     ,  /* 16  ++r16   */
                0x13L     ,  /* 17  ++r17   */
                0x14L     ,  /* 18  ++r18   */
                0x15L     ,  /* 19  ++r19   */
                0x18L     ,  /* 20  ++r20   */
                0x19L     ,  /* 21  ++r21   */
                0x1DL     ,  /* 22  ++r22   */

                0x16L     ,  /* 23  dec --  */
                0x17L     ,  /* 24  inc ++  */
                0x10L     ,  /* 25  no inc  */

                };







void CAU_0( int cond, int rH, unsigned long N) {


        /* Conditional branch group,
         *
         *    if ( !cond ) goto rH+N, return, ireturn, and nop.
         *
         */


        static unsigned long

                C[] = {         /* format CAU_0 cond C,G fields */


                        0x00L <<22 | 0x1L <<21,  /* 00  none        */

                        0x01L <<22 | 0x1L <<21,  /* 01  pl          */
                        0x01L <<22 | 0x0L <<21,  /* 02  mi          */
                        0x02L <<22 | 0x1L <<21,  /* 03  ne          */
                        0x02L <<22 | 0x0L <<21,  /* 04  eq          */
                        0x03L <<22 | 0x1L <<21,  /* 05  vc          */
                        0x03L <<22 | 0x0L <<21,  /* 06  vs          */
                        0x04L <<22 | 0x1L <<21,  /* 07  cc          */
                        0x04L <<22 | 0x0L <<21,  /* 08  cs          */
                        0x05L <<22 | 0x1L <<21,  /* 09  ge          */
                        0x05L <<22 | 0x0L <<21,  /* 10  lt          */
                        0x06L <<22 | 0x1L <<21,  /* 11  gt          */
                        0x06L <<22 | 0x0L <<21,  /* 12  le          */
                        0x07L <<22 | 0x1L <<21,  /* 13  hi          */
                        0x07L <<22 | 0x0L <<21,  /* 14  ls          */

                        0x08L <<22 | 0x1L <<21,  /* 15  auc         */
                        0x08L <<22 | 0x0L <<21,  /* 16  aus         */
                        0x09L <<22 | 0x1L <<21,  /* 17  age         */
                        0x09L <<22 | 0x0L <<21,  /* 18  alt         */
                        0x0AL <<22 | 0x1L <<21,  /* 19  ane         */
                        0x0AL <<22 | 0x0L <<21,  /* 20  aeq         */
                        0x0BL <<22 | 0x1L <<21,  /* 21  avc         */
                        0x0BL <<22 | 0x0L <<21,  /* 22  avs         */
                        0x0CL <<22 | 0x1L <<21,  /* 23  agt         */
                        0x0CL <<22 | 0x0L <<21,  /* 24  ale         */
                        0x0DL <<22 | 0x1L <<21,  /* 25  reserved    */
                        0x0DL <<22 | 0x0L <<21,  /* 26  reserved    */
                        0x0EL <<22 | 0x1L <<21,  /* 27  reserved    */
                        0x0EL <<22 | 0x0L <<21,  /* 28  reserved    */
                        0x0FL <<22 | 0x1L <<21,  /* 29  reserved    */
                        0x0FL <<22 | 0x0L <<21,  /* 30  reserved    */

                        0x10L <<22 | 0x1L <<21,  /* 31  ibe         */
                        0x10L <<22 | 0x0L <<21,  /* 32  ibf         */
                        0x11L <<22 | 0x1L <<21,  /* 33  obf         */
                        0x11L <<22 | 0x0L <<21,  /* 34  obe         */
                        0x12L <<22 | 0x1L <<21,  /* 35  pde         */
                        0x12L <<22 | 0x0L <<21,  /* 36  pdf         */
                        0x13L <<22 | 0x1L <<21,  /* 37  pie         */
                        0x13L <<22 | 0x0L <<21,  /* 38  pif         */
                        0x14L <<22 | 0x1L <<21,  /* 39  syc         */
                        0x14L <<22 | 0x0L <<21,  /* 40  sys         */
                        0x15L <<22 | 0x1L <<21,  /* 41  fbc         */
                        0x15L <<22 | 0x0L <<21,  /* 42  fbs         */
                        0x16L <<22 | 0x1L <<21,  /* 43  ireq1_lo    */
                        0x16L <<22 | 0x0L <<21,  /* 44  ireq1_hi    */
                        0x17L <<22 | 0x1L <<21,  /* 45  ireq2_lo    */
                        0x17L <<22 | 0x0L <<21,  /* 46  ireq2_hi    */

                        };


        check( &N, 0xFFFFL);

          emit( 0x00000000L |C[cond] |H[rH] |N);



}       
void CAU_2( void) {


        /* Reserved. */

        emit( 0x08000000L);


}
void CAU_3A( int rM, int rH, unsigned long N) {


        /* Loop counter, if ( rM-- >= 0 ) goto rH+N. */

        static unsigned long

                M[] = {

                        0x00L <<21,  /* 00  null    */

                        0x01L <<21,  /* 01  r1      */
                        0x02L <<21,  /* 02  r2      */
                        0x03L <<21,  /* 03  r3      */
                        0x04L <<21,  /* 04  r4      */
                        0x05L <<21,  /* 05  r5      */
                        0x06L <<21,  /* 06  r6      */
                        0x07L <<21,  /* 07  r7      */
                        0x08L <<21,  /* 08  r8      */
                        0x09L <<21,  /* 09  r9      */
                        0x0AL <<21,  /* 10  r10     */
                        0x0BL <<21,  /* 11  r11     */
                        0x0CL <<21,  /* 12  r12     */
                        0x0DL <<21,  /* 13  r13     */
                        0x0EL <<21,  /* 14  r14     */
                        0x11L <<21,  /* 15  r15     */
                        0x12L <<21,  /* 16  r16     */
                        0x13L <<21,  /* 17  r17     */
                        0x14L <<21,  /* 18  r18     */
                        0x15L <<21,  /* 19  r19     */
                        0x18L <<21,  /* 20  r20     */
                        0x19L <<21,  /* 21  r21     */
                        0x1DL <<21,  /* 22  r22     */ 

                        };



        check ( &N, 0xFFFFL);

          emit( 0x0C000000L |M[rM] |H[rH] |N);



}
void CAU_3B( unsigned long ninst, unsigned long NREPEAT) {



        /* Do immediate, do ninst, NREPEAT.
         *
         *    ninst+1 instructions are repeated NREPEAT+1 times ...
         *
         *
         */


        ninst <<= 16;

            emit( 0x8C000000L |ninst |NREPEAT );



}
void CAU_3C( unsigned long ninst, int rREPEAT) {



        /* Do register, do ninst, rREPEAT.
         *
         *    ninst+1 instructions are repeated rREPEAT+1 times ...
         *
         *
         */


        static unsigned long

                S[] = {

                        0x00L     ,  /* 00  null    */
                                
                        0x01L     ,  /* 01  r1      */
                        0x02L     ,  /* 02  r2      */
                        0x03L     ,  /* 03  r3      */
                        0x04L     ,  /* 04  r4      */
                        0x05L     ,  /* 05  r5      */
                        0x06L     ,  /* 06  r6      */
                        0x07L     ,  /* 07  r7      */
                        0x08L     ,  /* 08  r8      */
                        0x09L     ,  /* 09  r9      */
                        0x0AL     ,  /* 10  r10     */
                        0x0BL     ,  /* 11  r11     */
                        0x0CL     ,  /* 12  r12     */
                        0x0DL     ,  /* 13  r13     */
                        0x0EL     ,  /* 14  r14     */
                        0x11L     ,  /* 15  r15     */
                        0x12L     ,  /* 16  r16     */
                        0x13L     ,  /* 17  r17     */
                        0x14L     ,  /* 18  r18     */
                        0x15L     ,  /* 19  r19     */
                        0x18L     ,  /* 20  r20     */
                        0x19L     ,  /* 21  r21     */
                        0x1DL     ,  /* 22  r22     */

                        };


        ninst <<= 16;

           emit( 0x8C200000L |ninst |S[rREPEAT] );



}
void CAU_4( int rH, unsigned long N, int rM) {



        /* Call group, call rH+N (rM). */

        static unsigned long


                M[] = {

                        0x00L <<21,  /* 00  null    */

                        0x01L <<21,  /* 01  r1      */
                        0x02L <<21,  /* 02  r2      */
                        0x03L <<21,  /* 03  r3      */
                        0x04L <<21,  /* 04  r4      */
                        0x05L <<21,  /* 05  r5      */
                        0x06L <<21,  /* 06  r6      */
                        0x07L <<21,  /* 07  r7      */
                        0x08L <<21,  /* 08  r8      */
                        0x09L <<21,  /* 09  r9      */
                        0x0AL <<21,  /* 10  r10     */
                        0x0BL <<21,  /* 11  r11     */
                        0x0CL <<21,  /* 12  r12     */
                        0x0DL <<21,  /* 13  r13     */
                        0x0EL <<21,  /* 14  r14     */
                        0x11L <<21,  /* 15  r15     */
                        0x12L <<21,  /* 16  r16     */
                        0x13L <<21,  /* 17  r17     */
                        0x14L <<21,  /* 18  r18     */
                        0x15L <<21,  /* 19  r19     */
                        0x18L <<21,  /* 20  r20     */
                        0x19L <<21,  /* 21  r21     */
                        0x1DL <<21,  /* 22  r22     */

                        };



        check( &N, 0xFFFFL);

            emit( 0x10000000L |M[rM] |H[rH] |N);
        



}
void CAU_5A( int rD, int rH, unsigned long N) {


        /* Arithmetic/logic group, rD = rH + N, {ioc,dauc} = N. */


        static unsigned long

                D[] = {

                        0x1BL <<21,  /* 00  ioc     */
                        0x1AL <<21,  /* 01  dauc    */

                        };



        /* Check to make sure the register values aren't too long. */


        if ( rD == IOC )

                check( &N, 0x1FFFFFL); /* ioc, 21 bits */


        else if ( rD == DAUC )

                check( &N, 0x1FL); /* dauc, 5 bits */




        emit( 0x14000000L |D[rD] |H[rH] |N);



}
void CAU_5B( int rD, int rH, unsigned long N) {



        /* Arithmetic/logic group, 24 bit add, rD = rH + N. */


        static unsigned long

                D[] = {

                        0x00L <<21,  /* 00  null    */

                        0x01L <<21,  /* 01  r1      */
                        0x02L <<21,  /* 02  r2      */
                        0x03L <<21,  /* 03  r3      */
                        0x04L <<21,  /* 04  r4      */
                        0x05L <<21,  /* 05  r5      */
                        0x06L <<21,  /* 06  r6      */
                        0x07L <<21,  /* 07  r7      */
                        0x08L <<21,  /* 08  r8      */
                        0x09L <<21,  /* 09  r9      */
                        0x0AL <<21,  /* 10  r10     */
                        0x0BL <<21,  /* 11  r11     */
                        0x0CL <<21,  /* 12  r12     */
                        0x0DL <<21,  /* 13  r13     */
                        0x0EL <<21,  /* 14  r14     */
                        0x11L <<21,  /* 15  r15     */
                        0x12L <<21,  /* 16  r16     */
                        0x13L <<21,  /* 17  r17     */
                        0x14L <<21,  /* 18  r18     */
                        0x15L <<21,  /* 19  r19     */
                        0x18L <<21,  /* 20  r20     */
                        0x19L <<21,  /* 21  r21     */
                        0x1DL <<21,  /* 22  r22     */

                        };




        check( &N, 0xFFFFL);

            emit( 0x94000000L |D[rD] |H[rH] |N );



}       
void CAU_6B( int rD, int op, int rS, int func, int rT) {
        


        /* Arithmetic/logic group, 24 bit, rDe = rS F rT. */


        static unsigned long

                D[] = {

                        0x00L <<16,  /* 00  null    */

                        0x01L <<16,  /* 01  r1      */
                        0x02L <<16,  /* 02  r2      */
                        0x03L <<16,  /* 03  r3      */
                        0x04L <<16,  /* 04  r4      */
                        0x05L <<16,  /* 05  r5      */
                        0x06L <<16,  /* 06  r6      */
                        0x07L <<16,  /* 07  r7      */
                        0x08L <<16,  /* 08  r8      */
                        0x09L <<16,  /* 09  r9      */
                        0x0AL <<16,  /* 10  r10     */
                        0x0BL <<16,  /* 11  r11     */
                        0x0CL <<16,  /* 12  r12     */
                        0x0DL <<16,  /* 13  r13     */
                        0x0EL <<16,  /* 14  r14     */
                        0x11L <<16,  /* 15  r15     */
                        0x12L <<16,  /* 16  r16     */
                        0x13L <<16,  /* 17  r17     */
                        0x14L <<16,  /* 18  r18     */
                        0x15L <<16,  /* 19  r19     */
                        0x18L <<16,  /* 20  r20     */
                        0x19L <<16,  /* 21  r21     */
                        0x1DL <<16,  /* 22  r22     */

                        },

                E[] = {

                        0x00L <<11,  /* 00  two operand DYADIC      */
                        0x01L <<11,  /* 01  three operand TRIADIC   */

                        },


                T[] = {

                        0x10L << 5,  /* 00  zero    */

                        0x01L << 5,  /* 01  r1      */
                        0x02L << 5,  /* 02  r2      */
                        0x03L << 5,  /* 03  r3      */
                        0x04L << 5,  /* 04  r4      */
                        0x05L << 5,  /* 05  r5      */
                        0x06L << 5,  /* 06  r6      */
                        0x07L << 5,  /* 07  r7      */
                        0x08L << 5,  /* 08  r8      */
                        0x09L << 5,  /* 09  r9      */
                        0x0AL << 5,  /* 10  r10     */
                        0x0BL << 5,  /* 11  r11     */
                        0x0CL << 5,  /* 12  r12     */
                        0x0DL << 5,  /* 13  r13     */
                        0x0EL << 5,  /* 14  r14     */
                        0x11L << 5,  /* 15  r15     */
                        0x12L << 5,  /* 16  r16     */
                        0x13L << 5,  /* 17  r17     */
                        0x14L << 5,  /* 18  r18     */
                        0x15L << 5,  /* 19  r19     */
                        0x18L << 5,  /* 20  r20     */
                        0x19L << 5,  /* 21  r21     */
                        0x1DL << 5,  /* 22  r22     */

                        0x17L << 5,  /* 23  one     */

                        },


                S[] = {

                        0x00L     ,  /* 00  none    */

                        0x01L     ,  /* 01  r1      */
                        0x02L     ,  /* 02  r2      */
                        0x03L     ,  /* 03  r3      */
                        0x04L     ,  /* 04  r4      */
                        0x05L     ,  /* 05  r5      */
                        0x06L     ,  /* 06  r6      */
                        0x07L     ,  /* 07  r7      */
                        0x08L     ,  /* 08  r8      */
                        0x09L     ,  /* 09  r9      */
                        0x0AL     ,  /* 10  r10     */
                        0x0BL     ,  /* 11  r11     */
                        0x0CL     ,  /* 12  r12     */
                        0x0DL     ,  /* 13  r13     */
                        0x0EL     ,  /* 14  r14     */
                        0x11L     ,  /* 15  r15     */
                        0x12L     ,  /* 16  r16     */
                        0x13L     ,  /* 17  r17     */
                        0x14L     ,  /* 18  r18     */
                        0x15L     ,  /* 19  r19     */
                        0x18L     ,  /* 20  r20     */
                        0x19L     ,  /* 21  r21     */
                        0x1DL     ,  /* 22  r22     */

                        0x17L     ,  /* 23  one     */    

                        };



        emit( 0x98000000L |F[func] |D[rD] |E[op] |S[rS] |T[rT]);




}       
void CAU_6D( int rD, int func, unsigned long N) {
        

        /* Arithmetic/logic group, 24 bit, rDe = rD func N. */


        static unsigned long


                D[] = {

                        0x00L <<16,  /* 00  null    */

                        0x01L <<16,  /* 01  r1      */
                        0x02L <<16,  /* 02  r2      */
                        0x03L <<16,  /* 03  r3      */
                        0x04L <<16,  /* 04  r4      */
                        0x05L <<16,  /* 05  r5      */
                        0x06L <<16,  /* 06  r6      */
                        0x07L <<16,  /* 07  r7      */
                        0x08L <<16,  /* 08  r8      */
                        0x09L <<16,  /* 09  r9      */
                        0x0AL <<16,  /* 10  r10     */
                        0x0BL <<16,  /* 11  r11     */
                        0x0CL <<16,  /* 12  r12     */
                        0x0DL <<16,  /* 13  r13     */
                        0x0EL <<16,  /* 14  r14     */
                        0x11L <<16,  /* 15  r15     */
                        0x12L <<16,  /* 16  r16     */
                        0x13L <<16,  /* 17  r17     */
                        0x14L <<16,  /* 18  r18     */
                        0x15L <<16,  /* 19  r19     */
                        0x18L <<16,  /* 20  r20     */
                        0x19L <<16,  /* 21  r21     */
                        0x1DL <<16,  /* 22  r22     */

                        };



        check( &N, 0xFFFFL);

            emit( 0x9A000000L |F[func] |D[rD] |N);



}
void CAU_7A( int dir, int size, int rCAU, unsigned long N) {


        /* Data move group, CAU register <-> Memory direct *N. */


        unsigned long


                T[] = {

                        0x00L <<24,  /* 00  memory -> cau   */
                        0x01L <<24,  /* 01  memory <- cau   */

                        };


        
        check( &N, 0xFFFFL);

            emit( 0x1C000000L |T[dir] |W[size] |H[rCAU] |N);


        LOP();


}       
void CAU_7B( int dir, int size, int rCAU, MEM *m) {



        /* Data move group, CAU register <-> Memory indirect *rP++rIr. */


        static unsigned long

                T[] = {

                        0x00L <<24,  /* 00  memory -> cau   */
                        0x01L <<24,  /* 01  memory <- cau   */

                        };



        emit( 0x1E200000L |T[dir] |W[size] |H[rCAU]

                                        |r[m->rev] |P[m->rP] |I[m->rI]);

        LOP();



}
void CAU_7C( int dir, int size, int rIO, int rCAU ) {


        /* Data move group, IO register <-> CAU register. */


        static unsigned long


                T[] = {

                        0x00L <<24,  /* 00  cau -> io   */
                        0x01L <<24,  /* 01  cau <- io   */

                        },


                Wdefault[] = {

                        0x03L <<22,  /* 00  ibuf   long  */
                        0x03L <<22,  /* 01  obuf   long  */
                        0x02L <<22,  /* 02  pdr    word  */
                        0x01L <<22,  /* 03  piop   byte  */
                        0x02L <<22,  /* 04  pdr2   word  */
                        0x02L <<22,  /* 05  pir    word  */
                        0x02L <<22,  /* 06  pcw    word  */

                        },


                R[] = {

                        0x04L     ,  /* 00  ibuf    */
                        0x05L     ,  /* 01  obuf    */
                        0x06L     ,  /* 02  pdr     */
                        0x0EL     ,  /* 03  piop    */
                        0x14L     ,  /* 04  pdr2    */
                        0x16L     ,  /* 05  pir     */
                        0x1EL     ,  /* 06  pcw     */


                        };



        unsigned long w;

        
            w = ( size == DEFAULT ? Wdefault[rIO] : W[size] );



        emit( 0x1E000000L |T[dir] |w |H[rCAU] |R[rIO]);


        LOP();


}
void CAU_7D( int dir, int size, int rIO, MEM *m) {


        /* Data move group, IO register <-> Memory indirect *rP++rIr. */


        static unsigned long


                T[] = {

                        0x00L <<24,  /* 00  memory -> io   */
                        0x01L <<24,  /* 01  memory <- io   */

                        },


                Wdefault[] = {

                        0x03L <<22,  /* 00  ibuf   long  */
                        0x03L <<22,  /* 01  obuf   long  */
                        0x02L <<22,  /* 02  pdr    word  */
                        0x01L <<22,  /* 03  piop   byte  */
                        0x02L <<22,  /* 04  pdr2   long  */
                        0x02L <<22,  /* 05  pir    word  */
                        0x02L <<22,  /* 06  pcw    word  */

                        },


                R[] = {

                        0x04L <<16,  /* 00  ibuf    */
                        0x05L <<16,  /* 01  obuf    */
                        0x06L <<16,  /* 02  pdr     */
                        0x0EL <<16,  /* 03  piop    */
                        0x14L <<16,  /* 04  pdr2    */
                        0x16L <<16,  /* 05  pir     */
                        0x1EL <<16,  /* 06  pcw     */

                        };



        unsigned long w;


            w = ( size == DEFAULT ? Wdefault[rIO] : W[size] );




        emit( 0x1C200000L |T[dir] |w |R[rIO] |r[m->rev] |P[m->rP] |I[m->rI]);


        LOP();



}       
void CAU_8A( int rH, unsigned long M) {


        /* Unconditional branch group, goto rH+M. */


        static unsigned long

                NE, N;




        check( &M, 0xFFFFFFL);

          NE = (M&0x00FF0000L)<<5;
           N = (M&0x0000FFFFL);


        emit( 0xA0000000L |NE |H[rH] |N);



}
void CAU_8B( int rD, unsigned long M) {


        /* Arithmetic/logic group, rD = M. */


        static unsigned long

            NE, N,

                D[] = {

                        0x00L <<16,  /* 00  null    */

                        0x01L <<16,  /* 00  r1      */
                        0x02L <<16,  /* 01  r2      */
                        0x03L <<16,  /* 02  r3      */
                        0x04L <<16,  /* 03  r4      */
                        0x05L <<16,  /* 04  r5      */
                        0x06L <<16,  /* 05  r6      */
                        0x07L <<16,  /* 06  r7      */
                        0x08L <<16,  /* 07  r8      */
                        0x09L <<16,  /* 08  r9      */
                        0x0AL <<16,  /* 09  r10     */
                        0x0BL <<16,  /* 10  r11     */
                        0x0CL <<16,  /* 11  r12     */
                        0x0DL <<16,  /* 12  r13     */
                        0x0EL <<16,  /* 13  r14     */
                        0x11L <<16,  /* 14  r15     */
                        0x12L <<16,  /* 15  r16     */
                        0x13L <<16,  /* 16  r17     */
                        0x14L <<16,  /* 17  r18     */
                        0x15L <<16,  /* 18  r19     */
                        0x18L <<16,  /* 19  r20     */
                        0x19L <<16,  /* 20  r21     */
                        0x1DL <<16,  /* 21  r22     */

                        };



        check( &M, 0xFFFFFFL);

          NE = (M&0x00FF0000L)<<5;
           N = (M&0x0000FFFFL);


        emit( 0xC0000000L |NE |D[rD] |N);




}
void CAU_8C( unsigned long addr, int rM) {


        /* Call group, call M (rM). */

        static unsigned long

            NE, N,


                M[] = {

                        0x01L <<16,  /* 00  r1      */
                        0x02L <<16,  /* 01  r2      */
                        0x03L <<16,  /* 02  r3      */
                        0x04L <<16,  /* 03  r4      */
                        0x05L <<16,  /* 04  r5      */
                        0x06L <<16,  /* 05  r6      */
                        0x07L <<16,  /* 06  r7      */
                        0x08L <<16,  /* 07  r8      */
                        0x09L <<16,  /* 08  r9      */
                        0x0AL <<16,  /* 09  r10     */
                        0x0BL <<16,  /* 10  r11     */
                        0x0CL <<16,  /* 11  r12     */
                        0x0DL <<16,  /* 12  r13     */
                        0x0EL <<16,  /* 13  r14     */
                        0x11L <<16,  /* 14  r15     */
                        0x12L <<16,  /* 15  r16     */
                        0x13L <<16,  /* 16  r17     */
                        0x14L <<16,  /* 17  r18     */
                        0x15L <<16,  /* 18  r19     */
                        0x18L <<16,  /* 19  r20     */
                        0x19L <<16,  /* 20  r21     */
                        0x1DL <<16,  /* 21  r22     */

                        };




        check( &addr, 0xFFFFFFL);

          NE = (addr & 0x00FF0000L) <<5;
           N = (addr & 0x0000FFFFL);


        emit( 0xE0000000L |NE |M[rM-1] |N);



}
void R32_CALL_FIX( int return_reg) {


        /* DSP32C rev R32 call fix:
         *
         * The R32 revision of the DSP32C chip has a bug, always sign
         * extending a call return address from 16 to 24 bits.  A software
         * fix with no performance penelty is to emit r14 = . + 4 where the
         * latent nop would be normally emitted following the call.  This
         * fix is upward compatible with the R33 revision of the chip.
         *
         *
         */


        CAU_8B( return_reg, pc+4); /* return_reg = program location + 4 */

}








/* CAU CUSTOM EMITS:
 *
 *
 *
 *
 *
 */

void NOPS( int n) {


        /* Emit n nops. */

        while ( n-- )

                emit( 0x00000000L);


}
void LOP( void) {


        /* If nops are on, emit a cau latent nop. */

        if ( nops == 1 )

                emit( 0x00000000L);




}
int log2( unsigned long *N) {


        static unsigned long


                table[] = {

                        0x000001L,
                        0x000002L,
                        0x000004L,
                        0x000008L,

                        0x000010L,
                        0x000020L,
                        0x000040L,
                        0x000080L,

                        0x000100L,
                        0x000200L,
                        0x000400L,
                        0x000800L,

                        0x001000L,
                        0x002000L,
                        0x004000L,
                        0x008000L,

                        0x010000L,
                        0x020000L,
                        0x040000L,
                        0x080000L,

                        0x100000L,
                        0x200000L,
                        0x400000L,
                        0x800000L,

                        };

        unsigned i;



        /* Compute a base 2 logrithm with table look up. */

        for ( i = 0 ; i < 24 ; i++ ) {

                if ( *N == table[i] )

                        { *N = i; return( 1);}

                }



        return( 0); /* failure */




}
void ASHIFT( int rD, int func, int rS, unsigned long N) {


        /* Take the base 2 log of N. */

        if ( !log2( &N) )

                error( "N must be a power of 2, (0x2,0x4,...)");



        /* Emit rD = rS {*,/} 2, followed by rD {*=,/=} 2 */

        CAU_6B( rD, DYADIC, NONE, func, rS);

        while ( --N )

                CAU_6B( rD, DYADIC, NONE, func, rD);
        


}       
void LSHIFT( int rD, int func, int rS, unsigned long N) {



        /* Check to make sure N is reasonable. */

        if ( ! (1 <= N && N <= 0x17) )

                error( "N must be 0x1 <= N <= 0x17 (23 bits)");
        



        /* Emit rD = rS {>>,>>>,<<<} 1, followed by rD {>>=,>>>=,<<<=} 1 */

        CAU_6B( rD, DYADIC, NONE, func, rS);

        while ( --N )

                CAU_6B( rD, DYADIC, NONE, func, rD);
        



}       
void IFF_END( unsigned long cau_instruction, int cacond) {



        /* Fixup on iff ( cacond ) cau register expression. */


        static unsigned long

                C[] = {         /* format CAU_6B cacond C,G,K fields */


                        0x0L <<13 | 0x0L <<12 | 0x1L <<10,  /* 00  null  */

                        0x1L <<13 | 0x0L <<12 | 0x1L <<10,  /* 01  pl    */
                        0x1L <<13 | 0x1L <<12 | 0x1L <<10,  /* 02  mi    */
                        0x2L <<13 | 0x0L <<12 | 0x1L <<10,  /* 03  ne    */
                        0x2L <<13 | 0x1L <<12 | 0x1L <<10,  /* 04  eq    */
                        0x3L <<13 | 0x0L <<12 | 0x1L <<10,  /* 05  vc    */
                        0x3L <<13 | 0x1L <<12 | 0x1L <<10,  /* 06  vs    */
                        0x4L <<13 | 0x0L <<12 | 0x1L <<10,  /* 07  cc    */
                        0x4L <<13 | 0x1L <<12 | 0x1L <<10,  /* 08  cs    */
                        0x5L <<13 | 0x0L <<12 | 0x1L <<10,  /* 09  ge    */
                        0x5L <<13 | 0x1L <<12 | 0x1L <<10,  /* 10  lt    */
                        0x6L <<13 | 0x0L <<12 | 0x1L <<10,  /* 11  gt    */
                        0x6L <<13 | 0x1L <<12 | 0x1L <<10,  /* 12  le    */
                        0x7L <<13 | 0x0L <<12 | 0x1L <<10,  /* 13  hi    */
                        0x7L <<13 | 0x1L <<12 | 0x1L <<10,  /* 14  ls    */


                        };

        

        /* Check for a cau register CAU_6B format instruction. */

        if ( (cau_instruction&0xFE000000L) != 0x98000000L )

                error( "register arithmetic only with iff");
        



        /* Add the iff condition. */
        
        cau_instruction |= C[cacond];

            emit( cau_instruction);



}
void IF_BEGIN( int cond) {
        

        /* Emit if ( !cond ) goto pc+? at the top of the if block.
         *
         * NOTE:
         *
         * Keep in mind DSPASM uses positive logic, i.e. do the if block
         * if the condition is true, skip it if not true.
         *
         *
         */

        CAU_0( cond, PC, 0L);



}       
void IF_END( char SRFAR *if_fixup) {

        unsigned long N;



        /* Compute the offset from the beginning of the if to the bottom. */
        
        N = output - (if_fixup + 8);


        /* Fixup the if ( !cond ) goto pc+? at the top of the if block. */
        
        *(unsigned long SRFAR *)if_fixup |= N;



}
void ELSE_BEGIN( char SRFAR *if_fixup) {

        unsigned long N;



        /* Emit a goto pc+? at the bottom of the preceeding if block. */
        
        CAU_8A( PC, 0L);
        LOP();



        /* Fixup the if ( !cond ) goto pc+? at the top of the if block. */
        
        N = output - (if_fixup + 8);
        
        *(unsigned long SRFAR *)if_fixup |= N;
        


}
void ELSE_END( char SRFAR *else_fixup) {

        unsigned long N, NE;



        /* Fixup the goto pc+? at the bottom of the preceeding if block. */
        
        N = output - (else_fixup + 8);


        NE = (N&0x00FF0000L)<<5;
         N = (N&0x0000FFFFL);

        *(unsigned long SRFAR *)else_fixup |= (NE |N);



}
void WHILE_END( char SRFAR *top, char SRFAR *fixup) {

        unsigned long N;



        /* Emit a goto top at the bottom of the while loop, goto pc+? */

        N = (output + 8) - top;

        CAU_8A( PC, -N);
        LOP();



        /* Fixup the if ( !cond ) goto pc+? test instruction at the top. */

        N = output - (fixup + 8);
        
        *(unsigned long SRFAR *)fixup |= N;



}
void FOR_END( char SRFAR *top, char SRFAR *fixup, unsigned long inc) {

        unsigned long N;



        /* Emit a goto top at the bottom of the for loop, goto pc+?,
         * along with the incrementing statement on the following latent
         * instruction.
         *
         */


        N = (output + 8) - top;

        CAU_8A( PC, -N);
        emit( inc);




        /* Fixup the if ( !cond ) goto pc+? test instruction at the top. */

        N = output - (fixup + 8);
        
        *(unsigned long SRFAR *)fixup |= N;



}
void LOOP_END( char SRFAR *top, int rM) {

        unsigned long N;



        /* Emit if (rM-- >= 0) goto top at the bottom of the loop. */

        N = (output + 8) - top;


        if ( N > 0xFFFF )

                error( "offset to top of loop must be <= 0xFFFF");

/**** Add in offset checks like this ^^^^ to the if statements ??? ****/


 
        CAU_3A( rM, PC, -N);


}
void DO_REG_BEGIN( int rREPEAT) {


        /* Emit do ?,rS */

        CAU_3C( 0, rREPEAT);



}
void DO_IMM_BEGIN( unsigned long NREPEAT) {



        /* Special case NREPEAT == 0 for pass 1 of two pass assembly */

        if ( pflag == 2 && pass == 1 && NREPEAT == 0 )

                CAU_3B( 0, 0L);




        /* Emit do ?,NREPEAT */


        else if ( 1 <= NREPEAT && NREPEAT <= 0x800 )

                CAU_3B( 0, NREPEAT-1);

        else

                error( "repeat count N must be 0x1 <= N <= 0x800");



}
void DO_END( char SRFAR *fixup) {

        unsigned long N;



        /* Compute the number of instructions in the loop. */

        N = ( output - (fixup+4) )/4;



        /* Adjust by 1 for the J+1 instruction count. */

        if ( N > 0 )

                N--;

        else

                error( "too few instruction in do loop");



        /* Fixup the do instruction count at the top. */

        if ( N < 32 )

                *(unsigned long SRFAR *)fixup |= (N<<16);

        else

                error( "too many instructions in do loop");


}









/* DAU EMITS:
 *
 *
 * NOTE:
 *
 * The special cases of when Z and Y or Y and X would have the same rP
 * is not adequately documented in the AT&T DSP32C information manual.
 * In fact such instructions require the pbits field to be 1111.  Page
 * A-11 of the AT&T DSP32C information manual says this value is not
 * allowed.  However, IT IS used when the pbits fields would be the same.
 *
 * This encoding is refered to in the old AT&T DSP32 (no C) information
 * manual, and is followed by the AT&T DSP32C assembler.
 *
 * Additionally, this encoding is the only one executed correctly by the
 * DSP32C.  Encoding with the pbit fields the same does not execute properly.
 *
 *
 *
 *
 */

static unsigned long


        Fmt[] = {

                0x00000000L,  /* 00  null           */

                0x20000000L,  /* 01  format 1       */
                0x40000000L,  /* 02  format 2       */
                0x60000000L,  /* 03  format 3       */
                0x20000000L,  /* 04  format 4       */

                0x78000000L,  /* 05  format 5       */

                },


        Zp[] = {

                0x00L << 3,   /* 00  direct         */

                0x01L << 3,   /* 01  r1             */
                0x02L << 3,   /* 02  r2             */
                0x03L << 3,   /* 03  r3             */
                0x04L << 3,   /* 04  r4             */
                0x05L << 3,   /* 05  r5             */
                0x06L << 3,   /* 06  r6             */
                0x07L << 3,   /* 07  r7             */
                0x08L << 3,   /* 08  r8             */
                0x09L << 3,   /* 09  r9             */
                0x0AL << 3,   /* 10  r10            */
                0x0BL << 3,   /* 11  r11            */
                0x0CL << 3,   /* 12  r12            */
                0x0DL << 3,   /* 13  r13            */
                0x0EL << 3,   /* 14  r14            */

                0x0FL << 3,   /* 15  duplicate      */

                },


        Zi[] = {

                0x07L     ,   /* 00  no Z write     */

                0x00L     ,   /* 01  null           */
                0x00L     ,   /* 02  null           */
                0x00L     ,   /* 03  null           */
                0x00L     ,   /* 04  null           */

                0x05L     ,   /* 05  obuf           */
                0x06L     ,   /* 06  pdr            */

                0x00L     ,   /* 07  null           */
                0x00L     ,   /* 08  null           */
                0x00L     ,   /* 09  null           */
                0x00L     ,   /* 10  null           */
                0x00L     ,   /* 11  null           */
                0x00L     ,   /* 12  null           */
                0x00L     ,   /* 13  null           */
                0x00L     ,   /* 14  null           */

                0x01L     ,   /* 15  ++r15          */
                0x02L     ,   /* 16  ++r16          */
                0x03L     ,   /* 17  ++r17          */
                0x04L     ,   /* 18  ++r18          */
                0x05L     ,   /* 19  ++r19          */

                0x00L     ,   /* 20  null           */
                0x00L     ,   /* 21  null           */
                0x00L     ,   /* 22  null           */

                0x06L     ,   /* 23  dec --         */
                0x07L     ,   /* 24  inc ++         */
                0x00L     ,   /* 25  no inc         */

                },


        Yp[] = {

                0x00L <<10,   /* 00  direct         */

                0x01L <<10,   /* 01  r1             */
                0x02L <<10,   /* 02  r2             */
                0x03L <<10,   /* 03  r3             */
                0x04L <<10,   /* 04  r4             */
                0x05L <<10,   /* 05  r5             */
                0x06L <<10,   /* 06  r6             */
                0x07L <<10,   /* 07  r7             */
                0x08L <<10,   /* 08  r8             */
                0x09L <<10,   /* 09  r9             */
                0x0AL <<10,   /* 10  r10            */
                0x0BL <<10,   /* 11  r11            */
                0x0CL <<10,   /* 12  r12            */
                0x0DL <<10,   /* 13  r13            */
                0x0EL <<10,   /* 14  r14            */

                0x0FL <<10,   /* 15  duplicate      */

                },


        Yi[] = {

                0x00L << 7,   /* 00  a0             */
                0x01L << 7,   /* 01  a1             */
                0x02L << 7,   /* 02  a2             */
                0x03L << 7,   /* 03  a3             */

                0x04L << 7,   /* 04  ibuf           */

                0x00L << 7,   /* 05  null           */
                0x00L << 7,   /* 06  null           */
                0x00L << 7,   /* 07  null           */
                0x00L << 7,   /* 08  null           */
                0x00L << 7,   /* 09  null           */
                0x00L << 7,   /* 10  null           */
                0x00L << 7,   /* 11  null           */
                0x00L << 7,   /* 12  null           */
                0x00L << 7,   /* 13  null           */
                0x00L << 7,   /* 14  null           */

                0x01L << 7,   /* 15  ++r15          */
                0x02L << 7,   /* 16  ++r16          */
                0x03L << 7,   /* 17  ++r17          */
                0x04L << 7,   /* 18  ++r18          */
                0x05L << 7,   /* 19  ++r19          */

                0x00L << 7,   /* 20  null           */
                0x00L << 7,   /* 21  null           */
                0x00L << 7,   /* 22  null           */

                0x06L << 7,   /* 23  dec --         */
                0x07L << 7,   /* 24  inc ++         */
                0x00L << 7,   /* 25  no inc         */

                },


        N[] = {

                0x00L <<21,   /* 00  a0             */
                0x01L <<21,   /* 01  a1             */
                0x02L <<21,   /* 02  a2             */
                0x03L <<21,   /* 03  a3             */

                };


MEM

        NO_WRITE = { DIRECT,        0, 0 },
          Z_OBUF = { DIRECT, DAU_OBUF, 0 },
           Z_PDR = { DIRECT,  DAU_PDR, 0 },

           NO_XY = { DIRECT,        0, 0 };







void check_mem( MEM *m) {

        static char rI_error[] = {

                0,      /* 00  null       */

                1,      /* 01  ++r1       */
                1,      /* 02  ++r2       */
                1,      /* 03  ++r3       */
                1,      /* 04  ++r4       */
                1,      /* 05  ++r5       */
                1,      /* 06  ++r6       */
                1,      /* 07  ++r7       */
                1,      /* 08  ++r8       */
                1,      /* 09  ++r9       */
                1,      /* 10  ++r10      */
                1,      /* 11  ++r11      */
                1,      /* 12  ++r12      */
                1,      /* 13  ++r13      */
                1,      /* 14  ++r14      */

                0,      /* 15  ++r15      */
                0,      /* 16  ++r16      */
                0,      /* 17  ++r17      */
                0,      /* 18  ++r18      */
                0,      /* 19  ++r19      */

                1,      /* 20  ++r20      */
                1,      /* 21  ++r21      */
                1,      /* 22  ++r22      */

                0,      /* 23  post dec   */
                0,      /* 24  post inc   */
                0,      /* 25  no inc     */

                };




        /* Check a DAU memory reference for acceptable rP and rI. */

        if ( 15 <= m->rP && m->rP <= 22 )

                error( "all dau rP must be r1-r14");


        if ( m->rP != DIRECT && rI_error[ m->rI] )

                error( "all dau rI must be r15-r19");




}
void DAU_X( int fmt, int aM, int f, int s, int aN, MEM *X, MEM *Y, MEM *Z) {


        /* Generic DAU_X dau instruction format emit:
         *
         * The dau instruction formats 1-4 are all basically the same,
         * and are grouped together here as DAU_X.  Macro calls are defined
         * in dspasm.h to call this function appropriately for each format.
         *
         *
         */


        static unsigned long

                M[] = {

                        0x00L <<26,   /* 00  a0             */
                        0x01L <<26,   /* 01  a1             */
                        0x02L <<26,   /* 02  a2             */
                        0x03L <<26,   /* 03  a3             */

                        0x04L <<26,   /* 04  0.0 zero       */
                        0x05L <<26,   /* 05  1.0 one        */
                        0x06L <<26,   /* 06  format 4       */

                        },


                r[] = {

                        0x00L <<25,  /* 00  bit rev on      */
                        0x01L <<25,  /* 01  bit rev off     */

                        },


                F[] = {

                        0x00L <<24,  /* 00  + sign acc      */
                        0x01L <<24,  /* 01  - sign acc      */

                        },


                S[] = {

                        0x00L <<23,  /* 00  + sign mult     */
                        0x01L <<23,  /* 01  - sign mult     */

                        },


                Xp[] = {

                        0x00L <<17,  /* 00  direct          */

                        0x01L <<17,  /* 01  r1              */
                        0x02L <<17,  /* 02  r2              */
                        0x03L <<17,  /* 03  r3              */
                        0x04L <<17,  /* 04  r4              */
                        0x05L <<17,  /* 05  r5              */
                        0x06L <<17,  /* 06  r6              */
                        0x07L <<17,  /* 07  r7              */
                        0x08L <<17,  /* 08  r8              */
                        0x09L <<17,  /* 09  r9              */
                        0x0AL <<17,  /* 10  r10             */
                        0x0BL <<17,  /* 11  r11             */
                        0x0CL <<17,  /* 12  r12             */
                        0x0DL <<17,  /* 13  r13             */
                        0x0EL <<17,  /* 14  r14             */

                        0x0FL <<17,  /* 15  duplicate       */

                        },


                Xi[] = {

                        0x00L <<14,   /* 00  a0             */
                        0x01L <<14,   /* 01  a1             */
                        0x02L <<14,   /* 02  a2             */
                        0x03L <<14,   /* 03  a3             */

                        0x04L <<14,   /* 04  ibuf           */

                        0x00L <<14,   /* 05  null           */
                        0x00L <<14,   /* 06  null           */
                        0x00L <<14,   /* 07  null           */
                        0x00L <<14,   /* 08  null           */
                        0x00L <<14,   /* 09  null           */
                        0x00L <<14,   /* 10  null           */
                        0x00L <<14,   /* 11  null           */
                        0x00L <<14,   /* 12  null           */
                        0x00L <<14,   /* 13  null           */
                        0x00L <<14,   /* 14  null           */

                        0x01L <<14,   /* 15  ++r15          */
                        0x02L <<14,   /* 16  ++r16          */
                        0x03L <<14,   /* 17  ++r17          */
                        0x04L <<14,   /* 18  ++r18          */
                        0x05L <<14,   /* 19  ++r19          */

                        0x00L <<14,   /* 20  null           */
                        0x00L <<14,   /* 21  null           */
                        0x00L <<14,   /* 22  null           */

                        0x06L <<14,   /* 23  dec --         */
                        0x07L <<14,   /* 24  inc ++         */
                        0x00L <<14,   /* 25  no inc         */

                        };





        /* Check the memory references for r1-r14 and r15-r19. */

        check_mem( X);
        check_mem( Y);
        check_mem( Z);



        /* No reverse carry allowed on X or Y. */

        if ( X->rev || Y->rev )

                error( "no reverse carry allowed on X or Y");




        /* Special case Z->rP == Y->rP and Y->rP == X->rP. */

        if ( (Z->rP != DIRECT) && (Z->rP == Y->rP) )

                Z->rP = DUPLICATE;


        if ( (Y->rP != DIRECT) && (Y->rP == X->rP) )

                Y->rP = DUPLICATE;




        /* Emit the opcode. */
        
        emit( Fmt[fmt] |M[aM] |r[Z->rev] |F[f] |S[s] |N[aN]

                                        |Xp[X->rP] |Xi[X->rI]
                                        |Yp[Y->rP] |Yi[Y->rI]
                                        |Zp[Z->rP] |Zi[Z->rI]

                                        );



}
void DAU_5( MEM *Z, int aN, int func, MEM *Y) {


        /* Special function group, [Z] = aN = special( Y). */


        static unsigned long

                G[] = {

                        0x00L <<23,   /* 00 ic                */
                        0x01L <<23,   /* 01 oc                */
                        0x02L <<23,   /* 02 float             */
                        0x03L <<23,   /* 03 int               */
                        0x04L <<23,   /* 04 round             */
                        0x05L <<23,   /* 05 ifalt             */
                        0x06L <<23,   /* 06 ifaeq             */
                        0x07L <<23,   /* 07 ifagt             */
                        0x08L <<23,   /* 08 reserved          */
                        0x09L <<23,   /* 09 reserved          */
                        0x0AL <<23,   /* 10 float24           */
                        0x0BL <<23,   /* 11 int24             */
                        0x0CL <<23,   /* 12 ieee              */
                        0x0DL <<23,   /* 13 dsp               */
                        0x0EL <<23,   /* 14 seed              */
                        0x0FL <<23,   /* 15 reserved          */

                        };




        /* Check the memory references for r1-r14 and r15-r19. */

        check_mem( Y);
        check_mem( Z);



        /* No reverse carry allowed on Z or Y. */

        if ( Z->rev || Y->rev )

                error( "no reverse carry allowed on Z or Y");




        /* Special case Z->rP == Y->rP. */

        if ( (Z->rP != DIRECT) && (Z->rP == Y->rP) )

                Z->rP = DUPLICATE;




        /* Emit the opcode. */

        emit( Fmt[5] |G[func] |N[aN]

                                |Yp[Y->rP] |Yi[Y->rI]
                                |Zp[Z->rP] |Zi[Z->rI]

                                );

}
