/* FILE: dspio.c                Copyright (c), Symmetric Research, 1998-1999
 *
 * This is the Win9x version of the low level DSP32C support functions.  
 * The function calls in this file agree with those in the DOS and WinNT
 * dspio.c.  Applications that work with those files will also work with
 * this one without modification.
 *
 */




/* INCLUDE DSPIO.H:
 *
 * Note that for building the DLL library, the symbol _DLLEXPORT should be
 * defined on the compile command line.  See the makefile for details.
 *
 *
 *
 */

#include "dspio.h"







/* High and low word inline assembler helper macros:
 *
 * Access the high and low 16 bit words of a 32 bit long or pointer.
 *
 *
 *
 */

#define HW( p)          WORD PTR p[2]
#define LW( p)          WORD PTR p[0]







/* RELEASE REV INFO:
 *
 * Allow the user to determine the library rev they are working with.  See
 * dspio.h for more details.
 *
 */

VARTYPE( INT ) DSP32C_Rev = DSP32C_REV;

FUNCTYPE( INT ) DSP32C_GetRev( void ) {

        /* Return the library rev number. */

        return( DSP32C_Rev );
}







/* DEVICE DRIVER FUNCTIONS:
 *
 * These functions are NOPs provided for compatibility with systems that
 * need to open and close the communication connection between the DLL 
 * and a device driver.
 *
 */

FUNCTYPE( INT) DSP32C_Open( void) {
        return( 1 );
}
FUNCTYPE( INT) DSP32C_Close( void) {
        return( 1 );
}




/* PC IO SPACE BASE ADDRESS FUNCTIONS:
 *
 *
 *
 */

#define DEFAULT_BASEIO          0x280

VARTYPE( INT) DSP32C_BaseIo = 0x0000;


FUNCTYPE( INT) DSP32C_GetBaseIo( void) {

        return( DSP32C_BaseIo);

}
FUNCTYPE( VOID) DSP32C_SetBaseIo( unsigned NewBaseIo) {

        DSP32C_BaseIo = NewBaseIo;

}





/* DSPMUL CPU SELECTION:    (*** ONLY FOR DSPMUL MULTIPLE CPU BOARDS ***)
 *
 * For DSPMUL boards, which DSP32C cpu is currently connected to the ISA
 * bus is controlled by an additional on board PC IO space register.
 *
 * This byte wide register should be poked with the value of the desired
 * cpu.  DSP32C_SelectCpu MUST be called BEFORE DSP32C_Init to specify
 * which cpu will be initialized.  After initializing each cpu, it can
 * be called as needed to select a particular cpu for subsequent
 * operations.
 *
 *
 */

#define DEFAULT_BASESEL         0x2A0

VARTYPE( INT) DSP32C_BaseSel = 0x0000;


FUNCTYPE( INT) DSP32C_GetBaseSel( void) {

        return( DSP32C_BaseSel);

}
FUNCTYPE( VOID) DSP32C_SetBaseSel( unsigned NewBaseSel) {

        DSP32C_BaseSel = NewBaseSel;

}
FUNCTYPE( VOID) DSP32C_SelectCpu( unsigned n) {


        /* Set the DSPMUL cpu selection register. */

        _asm {

                mov     dx,LW( DSP32C_BaseSel)
                mov     ax,LW( n)
                out     dx,al

                }
}




/* DSP32C PIO REGISTER OFFSETS AND READ/WRITE FUNCTIONS:
 *
 *
 *
 *
 *
 *
 */

#define PCRl            (0x07*2)    /* DSP32 mode PCR low              */
#define PCRh            (0x0A*2)    /* DSP32C 8 bit mode PCR high      */

#define PAR             (0x00*2)    /* DSP32C 16 bit mode registers    */
#define PDR             (0x02*2)
#define EMR             (0x04*2)
#define ESR             (0x06*2)
#define PCR             (0x07*2)
#define PIR             (0x08*2)
#define PARE            (0x0B*2)
#define PDR2            (0x0C*2)


static int ReadPio( int reg_offset) {

        int value;


        /* Read a PIO register. */

        _asm {

                mov     eax,0

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,LW( reg_offset)
                in      ax,dx

                mov     value,eax

                }

        return( value);

}
static void WritePio( int reg_offset, int value) {


        /* Write a PIO register. */

        _asm {

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,LW( reg_offset)
                mov     ax,LW( value)
                out     dx,ax

                }
}








/* PROCESSOR INITIALIZATION:
 *
 * Note that DSP32C_BaseIo must be set before calling DSP32C_Init!
 *
 * In the past, setting DSP32C_BaseIo and DSP32C_BaseSel with DOS
 * environment space variables has been allowed.  This is supported here
 * with the C runtime getenv function.  THIS IS THE ONLY PLACE IN
 * DSPIO.C THAT THE C RUNTIME LIBRARY IS CALLED.
 *
 * If you wish to avoid the C runtime, compile without the calls to
 * getenv, and set up DSP32C_BaseIo and DSP32C_BaseSel on your own.
 *
 *
 *
 */

FUNCTYPE( INT) DSP32C_Init( void) {


#define _SEARCHENV

#ifdef _SEARCHENV

        char *EnvValue, *getenv( char *);

        /* Search for the DSP32C_BASEIO environment space variable. */

        if (DSP32C_BaseIo == 0x0000) {

                EnvValue = getenv( "DSP32C_BASEIO");

                if ( EnvValue )
                        sscanf( EnvValue, "%x", &DSP32C_BaseIo);
                else
                        DSP32C_BaseIo = DEFAULT_BASEIO;

                }



        /* Search for the DSP32C_BASESEL environement space variable. */

        if (DSP32C_BaseSel == 0x0000) {

                EnvValue = getenv( "DSP32C_BASESEL");

                if ( EnvValue )
                        sscanf( EnvValue, "%x", &DSP32C_BaseSel);
                else
                        DSP32C_BaseSel = DEFAULT_BASESEL;

                }


#else
        DSP32C_BaseIo  = DEFAULT_BASEIO;
        DSP32C_BaseSel = DEFAULT_BASESEL;

#endif




        /* Halt the DSP32C and configure the PIO port.
         *
         * The sequence of step are:
         *
         *   1.  set REGMAP for DSP32C PAB mode
         *   2.  convert to a 16 bit interface by setting PIO16
         *   3.  enable DMA and AUTO PIO data transfer modes
         *
         *
         */

        WritePio( PCRl, 0x02);         /* set REGMAP -> DSP32C PAB map */
        WritePio( PCRh, 0x02);         /* set PIO16 -> 16 bit interface */

            WritePio( PCR, 0x21A);         /* set DMA and AUTO */




        /* Do a presence test, checking the PCR register value. */

        if ( (ReadPio( PCR) & 0x71FL) != 0x21A )

                return( 0);



        /* Return success. */

        return( 1);


}










/* EXECUTION CONTROL:
 *
 *
 *
 */

FUNCTYPE( VOID) DSP32C_Go( void) {


        /* Reset the DSP32C and start execution from location 0.
         *
         * NOTE:
         *
         * As with DSP32C_Init, the PIO port must be reprogrammed for the 16
         * bit DMA mode after reset.
         *
         *
         */


        /* Toggle PCR[0], initiating reset and execution. */

        WritePio( PCRl, 0x00);
        WritePio( PCRl, 0x01);



        /* Reprogram the PIO port WHILE executing. */

        WritePio( PCRl, 0x03);         /* set REGMAP for DSP32C PAB map */
        WritePio( PCRh, 0x02);         /* set PIO16 for 16 bit interface */

            WritePio( PCR, 0x21B);     /* set DMA and AUTO */


}
FUNCTYPE( VOID) DSP32C_Halt( void) {

        int value;


        /* Halt the DSP32C by clearing PCR[0]. */

        value = ReadPio( PCR);

            value &= (~0x1);


        WritePio( PCR, value);

}










/* PIO REGISTER FUNCTIONS:
 *
 * These functions read and write the DSP32C PIO registers.
 *
 *
 *
 */

FUNCTYPE( INT) DSP32C_Pcr( void) {

        return( ReadPio( PCR));

}
FUNCTYPE( INT) DSP32C_Pir( void) {

        return( ReadPio( PIR));

}
FUNCTYPE( INT) DSP32C_Esr( void) {

        return( ReadPio( ESR));

}
FUNCTYPE( INT) DSP32C_Emr( void) {

        return( ReadPio( EMR));

}
FUNCTYPE( VOID) DSP32C_SetPir( unsigned value) {


        /* Set the PIR register to the specified value from the PC. */

        WritePio( PIR, value);


}
FUNCTYPE( VOID) DSP32C_SetEni( unsigned value) {

#define ENI     0x0004

        unsigned pcr;



        /* Set the value of the PCR[2] ENI bit.
         *
         * This bit enables or disables the DSP32C PIF pin which is connected
         * to the PC interrupt header on the board.  Value = 1 enables the
         * pin, value = 0 disables it.
         *
         * Note that the eni bit is automatically cleared by a reset of the
         * DSP32C, and must be reenabled after DSP32C_Init or Go if interrupts
         * are being used.
         *
         */

        pcr = ReadPio( PCR);

            if ( value == 1 )

                    pcr |= ENI;

            else
                    pcr &= ~ENI;



        WritePio( PCR, pcr);


}















/* DATA TRANSFER AND PROGRAM DOWNLOADING:
 *
 *
 *
 */

FUNCTYPE( VOID) DSP32C_Get( unsigned long src, void *dest, unsigned long nbytes) {


        /* Move data from the DSP32C to the PC, where:
         *
         *      src = DSP32C 24 bit source address
         *     dest = PC destination address
         *
         *
         */

        _asm {

                /* Point DSP32C (PARE,PAR) to src. */

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PARE
                mov     ax,HW( src)
                out     dx,ax

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PAR
                mov     ax,LW( src)
                out     dx,ax



                /* Rep insw DSP32C -> PC. */

                cld                             /* inc insw dest ...   */

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PDR

                mov     edi,dest
                mov     ecx,nbytes
                shr     ecx,1
                adc     ecx,0

                rep     insw

                }


}
FUNCTYPE( VOID) DSP32C_Put( void *src, unsigned long dest, unsigned long nbytes) {


        /* Move data from the PC to the DSP32C, where:
         *
         *      src = PC source address
         *     dest = DSP32C 24 bit destination address
         *
         *
         */

        _asm {

                /* Point DSP32C (PARE,PAR) to dest. */

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PARE
                mov     ax,HW( dest)
                out     dx,ax

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PAR
                mov     ax,LW( dest)
                out     dx,ax



                /* Rep outsw PC -> DSP32C PIO. */

                cld                             /* inc outsw src ...    */

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PDR

                mov     esi,src
                mov     ecx,nbytes
                shr     ecx,1
                adc     ecx,0

                rep     outsw

                }

}
FUNCTYPE( INT) DSP32C_SrHexLoad( unsigned long *input) {

        unsigned long dsp_start, nbytes;



        /* Download a DSPASM .dsh hex input array. */

        DSP32C_Halt();



        /* Check the leading array id. */

        if ( *input++ != 0x52535F5FL /* = "__SR */ )

                return( 0);



        /* Download records. */

        while ( *input != 0xFFFFFFFFL ) {

                dsp_start = *input++;
                   nbytes = *input++;

                   DSP32C_Put( (void *)input, dsp_start, nbytes);

                      ((char *)input) += nbytes;

                }



        /* Return success. */

        return( 1);

}







/* PUT LONG:
 *
 * This function writes a single 32 bit value to the DSP32C using the PIO
 * port 32 bit DMA mode.  The 16 bit PIO DMA autoinc mode generally used
 * by DSP32C_Get and Put is saved and restored on entry and exit.
 * 
 * This function is useful because it holds the PIO port off from
 * writing until it has a full 32 bits.  This allows the PC to poke
 * instructions and data onto the board while competing with DSP32C
 * execution, and not risk fetching problems as in the case of two 16
 * bit sequential writes.
 * 
 * The disadvantage of using this function is that it does not use the
 * outsw instruction like DSP32_Put.  In general, this function should
 * only be used for specialized applications, like breakpoint execution
 * in the DSPMON monitor.
 * 
 * As with the (PAR,PARE) registers, the order in which (PDR2,PDR) is
 * written is VERY important.  See the AT&T Information manual for
 * details.
 *
 *
 */

FUNCTYPE( VOID) DSP32C_PutLong( unsigned long value, unsigned long addr) {

        unsigned save, temp;



        /* Save the current DSP32C PCR value. */

        save = ReadPio( PCR);



        /* Program the PCR for a 32 bit DMA transfer. */

        temp = save;

            temp &= 0x0465;             /* mask off bits not to change */
            temp |= 0x031A;             /* set the 32 bit DMA mode */

            WritePio( PCR, temp);



        /* Transfer the long value. */

        _asm {

                /* (PARE,PAR) -> destination addr */

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PARE
                mov     ax,HW( addr)
                out     dx,ax

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PAR
                mov     ax,LW( addr)
                out     dx,ax


                /* Write the data through (PDR2,PDR). */

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PDR2
                mov     ax,LW( value)
                out     dx,ax

                mov     dx,LW( DSP32C_BaseIo)
                add     dx,PDR
                mov     ax,HW( value)
                out     dx,ax

                }



        /* Restore the PCR. */

        WritePio( PCR, save);

}

