/* FILE: add.dsp                Copyright (c), Symmetric Research, 1992-1998
 *
 * This simple DSPASM program adds two integer numbers together, saving
 * the result in the DSP32C memory.  Review some of the other programs
 * to see examples of more advanced DSPASM language features.
 *
 * To run it, assemble add.dsp with DSPASM, creating add.dso.  Then
 * download and execute add.dso by hand in the DSPMON monitor.  At the
 * DOS command line:
 *
 *              > dspasm add.dsp
 *              > dspmon add.dso
 *
 * where the filename extensions are required.
 *
 * Before running the program by hand in DSPMON, check to see that the
 * DSP32C memory location 100 is 0.  After running, it should be 5.
 *
 * For an example of a similar program written in the AT&T assembler,
 * see the file add.s.
 *
 *
 */




/* PROGRAM AND DATA LOCATION DEFINES:
 *
 * After the DSMON go command, the DSP32C execution begins at memory
 * location 0x000000.  Make sure your program begins there, or has a
 * goto statement to your entry point, so the DSP32C will have something
 * to execute.
 *
 * ALL constants in DSPASM are taken to be in hex.  A leading 0x is optional.
 *
 * This program has been written to be assembled in one pass.  No forward
 * references are used.
 *
 *
 *
 */

constant

        PROGRAM         = 0x000000,
        DATA            = 0x000100;






/* DATA AREA:
 *
 * Allocate the variables the program will use in this area.
 *
 *
 */

<DATA>

long sum = 0, x = 2, y = 3;





/* PROGRAM:
 *
 * The executable code is located in this area.
 *
 *
 *
 */

<PROGRAM>

 main {

        /* Add x and y, saving the result in sum. */

        r1 = *x;
        r2 = *y;

           r1 = r1 + r2;

        *sum = r1;


        /* Wait for the PC to come and get the result. */

        wait: goto wait;
}






/* MORE NOTES: 
 *
 * At the bottom of a program, the DSP32C will continue to execute off
 * into garbage instructions unless otherwise specified.  The easiest
 * way to deal with this is to finish off with an infinite goto loop,
 * having the DSP32C wait until the PC comes and gets the results.
 *
 * Also, in the program above, note that the direct addressing mode was
 * used in instructions like r1 = *x.  Be aware that because direct
 * addresses are encoded in the DSP32C instruction set with only 16 bits,
 * the label x MUST resolve to an address < 0x10000.  If DATA was changed
 * to a number greater than 16 bits, this would cause a problem.
 *
 * To access addresses greater than 0x10000, the indirect sequence r1 = x;
 * r1 = *r1 should be used.  Rewritten to use indirect addressing, the
 * above program would read as:
 * 
 * 
 *      r1 = x; r1 = *r1; r2 = y; r2 = *r2; r1 = r1 + r2;
 * 
 *      r3 = result; *r3 = r1;
 * 
 * 
 * DSPASM will generate an error messages for illegal direct addressing.
 * 
 * 
 */
