/* FILE: count.dsp              Copyright (c), Symmetric Research, 1992-1998
 *
 * This program shows how to use the DSPASM named register and
 * conditional loop constructs to write a program that looks very much
 * like C.  The program counts upward, saving the values in an array.
 *
 * To run it, assemble with DSPASM, then download and execute count.dso
 * by hand in the DSPMON monitor.  Check the values of Array1 and Array2
 * before and after the program is run.
 *
 *
 * NOTES:
 *
 * DSPASM's syntax for conditional and loop statements differs from that
 * of the AT&T/Lucent assembler.  In general, the body of the if, while,
 * or do construct is enclosed by curly braces {}, and treated as it
 * would be in C, where the curly braces are executed if the conditional
 * is TRUE.
 *
 * Also, DSPASM automatically generates nops for all latent CAU
 * instructions.  You can write CAU code and not worry about having to
 * insert nops.  NOPS ARE NOT required with the if, while, or do
 * statements.
 *
 * DSPASM does NOT automatically generate nops for DAU floating point
 * instructions.  You are responsible for properly using the DAU.
 *
 *
 *
 */

constant

        PROGRAM         = 0x000000,
        DATA            = 0x000100;




<DATA>

constant NARRAY = 10; /* <<< this is a hex 10, everything is hex in DSPASM ... */

long
        Array1[NARRAY],
        Array2[NARRAY];




<PROGRAM>

main {

        register
                        temp            = r1,
                        pArray          = r2,
                        n               = r3,
                        count           = r4;



        /* Test the if statement.  Note that {} is executed when TRUE. */

        temp = 2;

        if ( temp > 0 ) {


                /* Test the while statement, counting in Array1. */

                pArray = Array1;
                     n = NARRAY;
                 count = 1;

                while ( n-- > 0 )

                        { *pArray++ = count; count++; }
                

                /* Test the do statement, counting in Array2. */

                pArray = Array2;
                 count = 1;

                do NARRAY

                        { *pArray++ = count; count += 1; }

                }



        wait: goto wait;  /* wait for the PC ... */

}



/* MORE NOTES:
 *
 * Although the effect of while and do are similar in some respects, they
 * are executed differently from the DSP32C's point of view.
 *
 * The instructions in a while loop are execute from memory.  Those in a
 * do loop are kept in an internal instruction cache on the chip for
 * zero overhead looping.
 *
 * For do loops, ALL execution MUST occur from the instructions in the
 * do loop.  You cannot make subroutine calls or gotos from inside a do
 * loop, because it would break the chain of instructions in the cache.
 *
 * Do statements are particularly useful for blocks of code using the DAU
 * floating point multiply accumulate pipeline, effectively turning off
 * instruction fetching and increasing the throughput.
 *
 *
 *
 */
