/* FILE: diag.dsp               Copyright (c), Symmetric Research, 1996
 *
 * The DSP32C portion of the DSPA64 diag.exe program.
 *
 *
 *
 *
 *
 */

constant

        RESET           = 0x000000,

        SUCCESS         = 1,
        FAILED          = 2;





<RESET>

main {


        /* Main/reset entry point:
         *
         * The PC should poke this location with a goto for the required
         * function at reset.
         *
         *
         */

        goto wait_for_PC;

}







/* Test IBF:
 *
 * Test to make sure the IBF signal is being received.
 *
 *
 *
 *
 */

constant MAXCOUNT = 0x7000;


test_ibf {

        register

                value           = r1,
                signal          = r2,
                count           = r3;



        
        /* Initialize the serial port. */

        ioc = IN_MSB | ILEN_16 | OUT_MSB | OLEN_08;





        /* Test for IBF:
         *
         * Receiving IBF indicates the DSPA64 is connected to the DSPHLF
         * board correctly.
         *
         * The for loop times out if ibuf is never asserted.
         *
         *
         *
         */

        value = ibuf;    /* clear IBF ... and wait to get set ... */


        for ( count = 0 ; count < MAXCOUNT ; count++ ) {

                if ( ibf )

                        { signal = SUCCESS; pir = signal; goto wait_for_PC; }

                }




        /* Failed to get the IBF signal ... */

        signal = FAILED; pir = signal; goto wait_for_PC;


}





/* Acquire a value:
 *
 * Acquire data from given channel and returning it in VALUE.  N values
 * are averaged after clearing the acquisition pipeline.
 *
 *
 *
 */

constant N = 0x3E8;  /* = decimal 1000 */

word
        CHANNEL = 0,
          VALUE = 0;


dspf
        ZERO = 0.0, ONE_OVER_N = 0.001;  /* must be = 1/N */



acquire {


        register

                channel         = r1,
                value           = r2,
                count           = r3,
                signal          = r4,
                zero            = r5,
                one_over_n      = r6;




        /* Initialize the serial port. */

        ioc = IN_MSB | ILEN_16 | OUT_MSB | OLEN_08;




        /* Set the DSPA64 mux channel. */

        channel = *CHANNEL.w; obuf = channel.b;




        /* Acquire a value:
         *
         * Wait in a polling loop until ibf indicates a value.  Let the
         * acquisition pipeline clear before actually saving the value.
         *
         *
         */

             value = VALUE;
              zero = ZERO;
        one_over_n = ONE_OVER_N;

             count = (- 2);
                a1 = *zero;

   polling:

        if ( ibf ) {


                /* Convert the acquired value to floating point. */

                a0 = float( ibuf);

                    nop;
                    nop;
                    nop;



                /* Compute the average. */

                if ( count > 0 )

                        a1 = a1 + a0 * *one_over_n;


                if ( count == N ) {

                        *value = a1 = int( a1);

                          signal = SUCCESS; pir = signal; goto wait_for_PC;

                         }



                /* Increment the count. */

                count++;

                }


        goto polling;  /* continue polling */



}






/* Wait for the PC:
 *
 *
 *
 *
 */

wait_for_PC {


        /* Wait for the PC in a infinite loop. */

        wait: goto wait;


}








/* Test value:
 *
 * The PC reads this value to see if the DSPHLF was actually downloaded.
 *
 *
 *
 *
 *
 */

long TEST_VALUE = 0x11223344;













/* Serial port IOC control register bit assigments:
 *
 * These constants define the bit assigments for the serial port IOC
 * configuration control register.  They provide convenient mnemonic names
 * for the bits.
 *
 *
 */

constant

        ILEN_00         = 0x0 << 0x06,          /* input length         */
        ILEN_08         = 0x1 << 0x06, 
        ILEN_16         = 0x2 << 0x06,
        ILEN_32         = 0x3 << 0x06,

        OLEN_00         = 0x0 << 0x0A,          /* output length        */
        OLEN_08         = 0x1 << 0x0A,
        OLEN_16         = 0x2 << 0x0A,
        OLEN_32         = 0x3 << 0x0A,

        IN_LSB          = 0x0 << 0x10,          /* input lst first      */
        IN_MSB          = 0x1 << 0x10,          /* input msb first      */

        OUT_LSB         = 0x0 << 0x11,          /* output lsb first     */
        OUT_MSB         = 0x1 << 0x11;          /* output msb first     */

