// FILE: a64.h                Copyright (c), Symmetric Research, 2003
//
// Header file for DSPA64 specific portion of the Scope program includes
//
//      CA64App         main app    DSPA64 specific version of CScopeApp
//      CA64Win         main window DSPA64 specific version of CScopeWindow
//      CHardwareDlg    dialog for  DSPA64 hardware options
//      CChannelDlg     dialog for  DSPA64 channel options
//      CAcquireDlg     dialog for  DSPA64 acquisition options
//



// SPECIFIC RESOURCE CONSTANTS

#include "a64.rch"



// GENERAL APPLICATION CONSTANTS

#include "scope.h"



// SPECIFIC APPLICATION CONSTANTS

#define MAXCHKBTN       12


struct INI_DATAA64 {
        
        int   HardBaseio;
        int   HardJP60;
        float HardXtal;
        
        int   AcqNcount;
        int   AcqNormalize;
        int   AcqOverflow;
        float AcqCoeff;
        float AcqScale;
        
        };




//***********************************************************************



class CA64App : public CScopeApp {
        
public:

        // Data

        struct INI_DATAA64 IniA64;
        struct INI_DATAA64 IniA64Save;

        int    ReadyBuffer;
        float  BoardAggSPS;
        short *pDataArray;



        
        // Methods

        //{{AFX_VIRTUAL(CA64App)
        virtual int  ExitInstance();
        //}}AFX_VIRTUAL

        int ComputeTimingTicks();
        int ComputeSPSValues();
        int ComputeNcount();

        virtual CScopeWindow* NewScopeWindow();
        virtual int InitAppData();
        virtual int IniVerifyData();
        virtual int IniSetDefaults( void );
        virtual int IniSaveCurrent( void );
        virtual int IniRestoreCurrent( void );
        virtual int IniIsChanged( void );
        virtual int IniReadKeyword( void );
        virtual int IniWriteKeyword( FILE *OutFp );
        virtual int ComputeActiveChannels( void );
        virtual int AllocateDataArray( void );
        virtual int TriggerCondition( void );
        virtual int TriggerThresholdSetRange( int chan, long *thmin, long *thmax );
        virtual int TriggerThresholdCompare( int pt, long thmin, long thmax );
        virtual int TriggerDigioCompare( int line, int activestate );
        virtual int FillHdr( void );
        virtual int WriteBytes( CFile *outfile );

        virtual int AtodMeasure( int *ival );
        virtual int AtodInit( void );
        virtual int AtodStart( void );
        virtual int AtodCheckReady( void );
        virtual int AtodCheckOverflow( void );
        virtual int AtodGetData( void );
        virtual int AtodDemuxData( void );
        virtual int AtodStop( void );

        

        
        //{{AFX_MSG(CA64App)
        //}}AFX_MSG


        DECLARE_MESSAGE_MAP()
};      






//***********************************************************************


class CA64Win : public CScopeWindow {
        
public:

        
        // Methods

        CA64Win();
        ~CA64Win();

        
        //{{AFX_VIRTUAL(CA64Win)
        virtual void PostNcDestroy();
        //}}AFX_VIRTUAL

        
        //{{AFX_MSG(CA64Win)
        virtual afx_msg void OnHardwareDialog();
        virtual afx_msg void OnChannelDialog();
        virtual afx_msg void OnAcquireDialog();
        //}}AFX_MSG
        
        DECLARE_MESSAGE_MAP()
};






//***********************************************************************


class CHardwareDlg : public CScopeDialog {


public:
        CHardwareDlg(CWnd* pParent = NULL);   // standard constructor

        // Dialog Data
        
        //{{AFX_DATA(CHardwareDlg)
        enum { IDD = IDD_HARDWARE };
        CEdit   m_TXT_Maxrate;
        float   m_Maxrate;
        //}}AFX_DATA
        
        CEdit   m_TXT_Baseio;
        int     m_Baseio;
        int     m_Jumpers;
        int     m_Nticks;
        BOOL    m_NoChange;
        int     m_ChkBtn[MAXCHKBTN];
        int     m_LastFocusBtn;
        float   m_MinBoardAggSPS;
        float   m_MaxBoardAggSPS;



        // ClassWizard generated virtual function overrides
        
        //{{AFX_VIRTUAL(CHardwareDlg)
        protected:
        virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
        //}}AFX_VIRTUAL


protected:

        void UpdateJumper();
        void ComputeNticks();
        void ComputeMaxrate();
        void SetJumperBits();


        // Generated message map functions
        
        //{{AFX_MSG(CHardwareDlg)
        virtual BOOL OnInitDialog();
        virtual void OnCancel();
        virtual void OnOK();
        virtual LONG OnEditLostFocus( UINT nID, LONG xxx );
        afx_msg void OnHelp();
        afx_msg void OnSetfocusBits();
        afx_msg void OnKillfocusBits();
        afx_msg void OnChangeJumperBits();
        afx_msg void OnKillfocusBaseio();
        afx_msg void OnKillfocusMaxrate();
        //}}AFX_MSG
        
        DECLARE_MESSAGE_MAP()
};





//***********************************************************************





class CChannelDlg : public CScopeDialog {
        

public:
        CChannelDlg(CWnd* pParent = NULL);   // standard constructor
        ~CChannelDlg();                      // destructor

        
        // Data
        
        int  m_Nscan;
        int  *pTempActive;
        char *pTempTitle;
        
        //{{AFX_DATA(CChannelDlg)
        enum { IDD = IDD_CHANNEL };
        CListBox  m_LSTscan;
        int       m_CurrChan;
        //}}AFX_DATA


        // ClassWizard generated virtual function overrides
        
        //{{AFX_VIRTUAL(CChannelDlg)
        protected:
        virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
        //}}AFX_VIRTUAL


protected:

        void UpdateChannelData( int icurr );
        void UpdateInfo();
        void AddItem( int ikey );
        int  FindPos();
        

        // Generated message map functions
        
        //{{AFX_MSG(CChannelDlg)
        virtual BOOL OnInitDialog();
        virtual void OnCancel();
        virtual void OnOK();
        virtual LONG OnEditLostFocus( UINT nID, LONG xxx );
        afx_msg void OnHelp();
        afx_msg void OnBtnBack();
        afx_msg void OnBtnNext();
        afx_msg void OnBtnAllOn();
        afx_msg void OnBtnAllOff();
        afx_msg void OnChkOnOff();
        afx_msg void OnChangeTxtCurr();
        afx_msg void OnKillfocusTxtCurr();
        afx_msg void OnKillfocusTxtTitle();
        afx_msg void OnSelchangeListScan();
        //}}AFX_MSG
        
        DECLARE_MESSAGE_MAP()
};





//***********************************************************************

class CAcquireDlg : public CScopeDialog {
        

public:
        CAcquireDlg(CWnd* pParent = NULL);   // standard constructor

        // Dialog Data
        
        //{{AFX_DATA(CAcquireDlg)
        enum { IDD = IDD_ACQUIRE };
        int     m_DSPenable;
        int     m_Normalize;
        int     m_Overflow;
        float   m_Coeff;
        float   m_Scale;
        float   m_SPS;
        int     m_Ncount;
        //}}AFX_DATA


        // ClassWizard generated virtual function overrides
        
        //{{AFX_VIRTUAL(CAcquireDlg)
        protected:
        virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
        //}}AFX_VIRTUAL


protected:

        void ComputeSampleRate();
        void ComputeScale();


        // Generated message map functions
        
        //{{AFX_MSG(CAcquireDlg)
        virtual BOOL OnInitDialog();
        virtual void OnCancel();
        virtual void OnOK();
        virtual LONG OnEditLostFocus( UINT nID, LONG xxx );
        afx_msg void OnHelp();
        afx_msg void OnBtnChannels();
        afx_msg void OnChkDSPenable();
        afx_msg void OnChkNormalize();
        afx_msg void OnChkOverflow();
        afx_msg void OnKillfocusTxtCoeff();
        afx_msg void OnKillfocusTxtNcount();
        afx_msg void OnKillfocusTxtScale();
        afx_msg void OnKillfocusTxtSPS();
        //}}AFX_MSG
        
        DECLARE_MESSAGE_MAP()
};
